/*
 * Generated by gdbus-codegen 2.50.0. DO NOT EDIT.
 *
 * The license of this code is the same as for the source it was derived from.
 */

#ifndef __GF_FD_NOTIFICATIONS_GEN_H__
#define __GF_FD_NOTIFICATIONS_GEN_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.Notifications */

#define GF_TYPE_FD_NOTIFICATIONS_GEN (gf_fd_notifications_gen_get_type ())
#define GF_FD_NOTIFICATIONS_GEN(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GF_TYPE_FD_NOTIFICATIONS_GEN, GfFdNotificationsGen))
#define GF_IS_FD_NOTIFICATIONS_GEN(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GF_TYPE_FD_NOTIFICATIONS_GEN))
#define GF_FD_NOTIFICATIONS_GEN_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), GF_TYPE_FD_NOTIFICATIONS_GEN, GfFdNotificationsGenIface))

struct _GfFdNotificationsGen;
typedef struct _GfFdNotificationsGen GfFdNotificationsGen;
typedef struct _GfFdNotificationsGenIface GfFdNotificationsGenIface;

struct _GfFdNotificationsGenIface
{
  GTypeInterface parent_iface;


  gboolean (*handle_close_notification) (
    GfFdNotificationsGen *object,
    GDBusMethodInvocation *invocation,
    guint arg_id);

  gboolean (*handle_get_capabilities) (
    GfFdNotificationsGen *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_get_server_information) (
    GfFdNotificationsGen *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_notify) (
    GfFdNotificationsGen *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_app_name,
    guint arg_replaces_id,
    const gchar *arg_app_icon,
    const gchar *arg_summary,
    const gchar *arg_body,
    const gchar *const *arg_actions,
    GVariant *arg_hints,
    gint arg_expire_timeout);

  void (*action_invoked) (
    GfFdNotificationsGen *object,
    guint arg_id,
    const gchar *arg_action_key);

  void (*notification_closed) (
    GfFdNotificationsGen *object,
    guint arg_id,
    guint arg_reason);

};

GType gf_fd_notifications_gen_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *gf_fd_notifications_gen_interface_info (void);
guint gf_fd_notifications_gen_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void gf_fd_notifications_gen_complete_close_notification (
    GfFdNotificationsGen *object,
    GDBusMethodInvocation *invocation);

void gf_fd_notifications_gen_complete_get_capabilities (
    GfFdNotificationsGen *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *capabilities);

void gf_fd_notifications_gen_complete_get_server_information (
    GfFdNotificationsGen *object,
    GDBusMethodInvocation *invocation,
    const gchar *name,
    const gchar *vendor,
    const gchar *version,
    const gchar *spec_version);

void gf_fd_notifications_gen_complete_notify (
    GfFdNotificationsGen *object,
    GDBusMethodInvocation *invocation,
    guint id);



/* D-Bus signal emissions functions: */
void gf_fd_notifications_gen_emit_action_invoked (
    GfFdNotificationsGen *object,
    guint arg_id,
    const gchar *arg_action_key);

void gf_fd_notifications_gen_emit_notification_closed (
    GfFdNotificationsGen *object,
    guint arg_id,
    guint arg_reason);



/* D-Bus method calls: */
void gf_fd_notifications_gen_call_close_notification (
    GfFdNotificationsGen *proxy,
    guint arg_id,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gf_fd_notifications_gen_call_close_notification_finish (
    GfFdNotificationsGen *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gf_fd_notifications_gen_call_close_notification_sync (
    GfFdNotificationsGen *proxy,
    guint arg_id,
    GCancellable *cancellable,
    GError **error);

void gf_fd_notifications_gen_call_get_capabilities (
    GfFdNotificationsGen *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gf_fd_notifications_gen_call_get_capabilities_finish (
    GfFdNotificationsGen *proxy,
    gchar ***out_capabilities,
    GAsyncResult *res,
    GError **error);

gboolean gf_fd_notifications_gen_call_get_capabilities_sync (
    GfFdNotificationsGen *proxy,
    gchar ***out_capabilities,
    GCancellable *cancellable,
    GError **error);

void gf_fd_notifications_gen_call_get_server_information (
    GfFdNotificationsGen *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gf_fd_notifications_gen_call_get_server_information_finish (
    GfFdNotificationsGen *proxy,
    gchar **out_name,
    gchar **out_vendor,
    gchar **out_version,
    gchar **out_spec_version,
    GAsyncResult *res,
    GError **error);

gboolean gf_fd_notifications_gen_call_get_server_information_sync (
    GfFdNotificationsGen *proxy,
    gchar **out_name,
    gchar **out_vendor,
    gchar **out_version,
    gchar **out_spec_version,
    GCancellable *cancellable,
    GError **error);

void gf_fd_notifications_gen_call_notify (
    GfFdNotificationsGen *proxy,
    const gchar *arg_app_name,
    guint arg_replaces_id,
    const gchar *arg_app_icon,
    const gchar *arg_summary,
    const gchar *arg_body,
    const gchar *const *arg_actions,
    GVariant *arg_hints,
    gint arg_expire_timeout,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gf_fd_notifications_gen_call_notify_finish (
    GfFdNotificationsGen *proxy,
    guint *out_id,
    GAsyncResult *res,
    GError **error);

gboolean gf_fd_notifications_gen_call_notify_sync (
    GfFdNotificationsGen *proxy,
    const gchar *arg_app_name,
    guint arg_replaces_id,
    const gchar *arg_app_icon,
    const gchar *arg_summary,
    const gchar *arg_body,
    const gchar *const *arg_actions,
    GVariant *arg_hints,
    gint arg_expire_timeout,
    guint *out_id,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define GF_TYPE_FD_NOTIFICATIONS_GEN_PROXY (gf_fd_notifications_gen_proxy_get_type ())
#define GF_FD_NOTIFICATIONS_GEN_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GF_TYPE_FD_NOTIFICATIONS_GEN_PROXY, GfFdNotificationsGenProxy))
#define GF_FD_NOTIFICATIONS_GEN_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GF_TYPE_FD_NOTIFICATIONS_GEN_PROXY, GfFdNotificationsGenProxyClass))
#define GF_FD_NOTIFICATIONS_GEN_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GF_TYPE_FD_NOTIFICATIONS_GEN_PROXY, GfFdNotificationsGenProxyClass))
#define GF_IS_FD_NOTIFICATIONS_GEN_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GF_TYPE_FD_NOTIFICATIONS_GEN_PROXY))
#define GF_IS_FD_NOTIFICATIONS_GEN_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GF_TYPE_FD_NOTIFICATIONS_GEN_PROXY))

typedef struct _GfFdNotificationsGenProxy GfFdNotificationsGenProxy;
typedef struct _GfFdNotificationsGenProxyClass GfFdNotificationsGenProxyClass;
typedef struct _GfFdNotificationsGenProxyPrivate GfFdNotificationsGenProxyPrivate;

struct _GfFdNotificationsGenProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  GfFdNotificationsGenProxyPrivate *priv;
};

struct _GfFdNotificationsGenProxyClass
{
  GDBusProxyClass parent_class;
};

GType gf_fd_notifications_gen_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GfFdNotificationsGenProxy, g_object_unref)
#endif

void gf_fd_notifications_gen_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GfFdNotificationsGen *gf_fd_notifications_gen_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
GfFdNotificationsGen *gf_fd_notifications_gen_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void gf_fd_notifications_gen_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GfFdNotificationsGen *gf_fd_notifications_gen_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
GfFdNotificationsGen *gf_fd_notifications_gen_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define GF_TYPE_FD_NOTIFICATIONS_GEN_SKELETON (gf_fd_notifications_gen_skeleton_get_type ())
#define GF_FD_NOTIFICATIONS_GEN_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GF_TYPE_FD_NOTIFICATIONS_GEN_SKELETON, GfFdNotificationsGenSkeleton))
#define GF_FD_NOTIFICATIONS_GEN_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GF_TYPE_FD_NOTIFICATIONS_GEN_SKELETON, GfFdNotificationsGenSkeletonClass))
#define GF_FD_NOTIFICATIONS_GEN_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GF_TYPE_FD_NOTIFICATIONS_GEN_SKELETON, GfFdNotificationsGenSkeletonClass))
#define GF_IS_FD_NOTIFICATIONS_GEN_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GF_TYPE_FD_NOTIFICATIONS_GEN_SKELETON))
#define GF_IS_FD_NOTIFICATIONS_GEN_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GF_TYPE_FD_NOTIFICATIONS_GEN_SKELETON))

typedef struct _GfFdNotificationsGenSkeleton GfFdNotificationsGenSkeleton;
typedef struct _GfFdNotificationsGenSkeletonClass GfFdNotificationsGenSkeletonClass;
typedef struct _GfFdNotificationsGenSkeletonPrivate GfFdNotificationsGenSkeletonPrivate;

struct _GfFdNotificationsGenSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  GfFdNotificationsGenSkeletonPrivate *priv;
};

struct _GfFdNotificationsGenSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType gf_fd_notifications_gen_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GfFdNotificationsGenSkeleton, g_object_unref)
#endif

GfFdNotificationsGen *gf_fd_notifications_gen_skeleton_new (void);


G_END_DECLS

#endif /* __GF_FD_NOTIFICATIONS_GEN_H__ */
